package org.implab.gradle.containers.cli;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.Set;

import org.gradle.api.logging.Logger;

public abstract class DockerTraits {

    public final String BUILD_COMMAND = "build";
    public final String PUSH_COMMAND = "push";
    public final String RUN_COMMAND = "run";
    public final String SAVE_COMMAND = "save";
    public final String INSPECT_COMMAND = "inspect";
    public final String IMAGE_COMMAND = "image";
    public final String TAG_COMMAND = "tag";
    public final String COMPOSE_COMMAND = "compose";
    public final String UP_COMMAND = "up";
    public final String STOP_COMMAND = "stop";
    public final String RM_COMMAND = "rm";

    public abstract Logger getLogger();

    public abstract Optional<File> getWorkingDir();

    public abstract String getCliCmd();

    Process startProcess(ProcessBuilder builder) throws IOException {
        getLogger().info("Starting: {}", builder.command());
        return builder.start();
    }

    protected boolean checkRetCode(Process proc, int code) throws InterruptedException {
        if (getLogger().isInfoEnabled()) {
            Utils.redirectIO(proc.getInputStream(), getLogger()::info);
            Utils.redirectIO(proc.getErrorStream(), getLogger()::info);
        }

        return proc.waitFor() == code;
    }

    protected void complete(Process proc) throws InterruptedException, IOException {
        if (getLogger().isInfoEnabled())
            Utils.redirectIO(proc.getInputStream(), getLogger()::info);

        if (getLogger().isErrorEnabled())
            Utils.redirectIO(proc.getErrorStream(), getLogger()::error);

        var code = proc.waitFor();
        if (code != 0) {
            getLogger().error("The process exited with code {}", code);
            throw new IOException("The process exited with error code " + code);
        }
    }

    protected ProcessBuilder builder(String... args) {
        var argsList = new ArrayList<String>(args.length + 1);
        Arrays.stream(args).forEach(argsList::add);

        return builder(argsList);
    }

    protected ProcessBuilder builder(List<String> args) {
        var command = new ArrayList<String>(args.size() + 1);

        command.add(getCliCmd());
        args.forEach(command::add);

        var builder = new ProcessBuilder(command);

        getWorkingDir().ifPresent(builder::directory);
        return builder;
    }

    public void buildImage(String imageName, File contextDirectory, List<String> options)
            throws IOException, InterruptedException {
        var args = new ArrayList<String>();
        args.add(BUILD_COMMAND);
        args.addAll(options);
        args.add("-t");
        args.add(imageName);
        args.add(contextDirectory.getAbsolutePath());
        complete(startProcess(builder(args)));
    }

    public void pushImage(String image, List<String> options) throws InterruptedException, IOException {
        var args = new ArrayList<String>();
        args.add(PUSH_COMMAND);
        args.addAll(options);
        args.add(image);
        complete(startProcess(builder(args)));
    }

    public void runImage(String image, List<String> options, List<String> command)
            throws InterruptedException, IOException {
        var args = new ArrayList<String>();
        args.add(RUN_COMMAND);
        args.addAll(options);
        args.add(image);
        args.addAll(command);
        complete(startProcess(builder(args)));
    }

    public void saveImage(List<String> images, File output) throws InterruptedException, IOException {
        if (output.exists())
            output.delete();

        var args = new ArrayList<String>();
        args.add(SAVE_COMMAND);
        args.add("-o");
        args.add(output.getAbsolutePath());
        images.forEach(args::add);

        complete(startProcess(builder(args)));
    }

    public void tagImage(String source, String target) throws InterruptedException, IOException {
        complete(startProcess(builder(TAG_COMMAND, source, target)));
    }

    public boolean imageExists(String imageId) throws InterruptedException, IOException {
        getLogger().info("Check image {} exists", imageId);

        return checkRetCode(
                startProcess(builder(IMAGE_COMMAND, INSPECT_COMMAND, "--format", "image-exists", imageId)),
                0);
    }

    public boolean imageExists(File imageIdFile) {
        if (imageIdFile.exists()) {
            try {
                var imageId = Utils.readImageRef(imageIdFile);
                return imageExists(imageId.getId());
            } catch (IOException | InterruptedException e) {
                getLogger().error("Failed to read imageId {}: {}", imageIdFile, e);
                return false;
            }
        }
        return false;
    }

    List<String> composeArgs(File primaryCompose, Set<String> profiles, String... extra) {
        var args = new ArrayList<String>();

        args.add(COMPOSE_COMMAND);
        args.add("--file");
        args.add(primaryCompose.getAbsolutePath());

        if (profiles.size() > 0) {
            for (var profile : profiles) {
                args.add("--profile");
                args.add(profile);
            }
        }

        args.addAll(List.of(extra));

        return args;
    }

    public void composeUp(File primaryCompose, Set<String> profiles) throws InterruptedException, IOException {
        var args = composeArgs(primaryCompose, profiles, UP_COMMAND, "--detach");

        complete(startProcess(builder(args)));
    }

    public void composeStop(File primaryCompose, Set<String> profiles) throws InterruptedException, IOException {
        var args = composeArgs(primaryCompose, profiles, STOP_COMMAND);
        complete(startProcess(builder(args)));
    }

    public void composeRm(File primaryCompose, Set<String> profiles, boolean removeVolumes)
            throws InterruptedException, IOException {
        var args = composeArgs(primaryCompose, profiles, RM_COMMAND, "--force", "--stop");

        if (removeVolumes)
            args.add("--volumes");

        complete(startProcess(builder(args)));
    }
}
